"""
Sastre - Automation Tools for Cisco SD-WAN Powered by Viptela

"""
__copyright__ = "Copyright (c) 2019-2022 Cisco Systems, Inc. and/or its affiliates"
__version__   = "1.18.1"
__author__    = "Marcelo Reis"
__email__     = "sastre-support@cisco.com"
__url__       = "https://github.com/CiscoDevNet/sastre"
